"""

Routines for setting up and performing training

"""
import sys
import collections
from dataclasses import dataclass
import copy
import timeit

from typing import Callable, Union, Optional
import torch

import pickle

from . import serialization
from .metric_tracker import MetricTracker
from .controllers import Controller, is_scheduler_like
from .device import set_devices
from .. import tools
from .assembly import TrainingModules
from .step_functions import get_step_function

from .. import custom_kernels


@dataclass
class SetupParams:
    """
    :param stopping_key:  name of validation metric for stopping
    :param controller: Optional -- Controller object for LR scheduling and ending experiment.
                        If not provided, will be constructed from parameters below.
    :param Device: Where to train the model.
                   Falls back to CUDA if available.
                   Specify a tuple of device numbers to use DataParallel.
    :param max_epochs: Optional -- maximum number of epochs to train.
                    Mandatory if controller not provided.
    :param batch_size:
                    Only used if the controller itself is not specified.
                    Mandatory if controller not provided.
    :param eval_batch_size:
                    Only used if the controller itself is not specified.
    :param scheduler: scheduler passed to the controller
    :param optimizer: Pytorch optimizer or optimizer class. Defaults to Adam.
    :param learning_rate: If an optimizer class is provided, the learning rate is used to construct the optimizer.
    :param fraction_train_eval: What fraction of the training dataset to evaluate in the evaluation phase

    All params after stopping_key, controller, and device are optional and can be built into a controller.

    .. Note::

        Multiple GPUs is an experimental feature that is currently under debugging.
    """

    device: Optional[Union[torch.device, str]] = None
    controller: Optional[Union[Controller, Callable]] = None
    stopping_key: Optional[str] = None
    optimizer: Union[torch.optim.Optimizer, Callable] = torch.optim.Adam
    learning_rate: Optional[float] = None
    scheduler: Union[torch.optim.lr_scheduler._LRScheduler, Callable] = None
    batch_size: Optional[int] = None
    eval_batch_size: Optional[int] = None
    max_epochs: Optional[int] = None
    fraction_train_eval: Optional[float] = 0.1

    _controller_params = (
        "stopping_key",
        "optimizer",
        "learning_rate",
        "scheduler",
        "batch_size",
        "eval_batch_size",
        "max_epochs",
        "fraction_train_eval",
    )

    def __post_init__(self):
        if self.eval_batch_size is None:
            self.eval_batch_size = self.batch_size
        if isinstance(self.controller, Controller):
            # Warn if extra arguments are specified, but only if they are not default values.
            for param_str in self._controller_params:
                pval = getattr(self, param_str)
                classpval = getattr(self.__class__, param_str)
                if pval is not None and pval is not classpval:
                    print(
                        "Warning: When controller is specified, argument "
                        f"'{param_str}' should be given to the controller. "
                        f"Value ({getattr(self,param_str)}) ignored.",
                        file=sys.stderr,
                    )
        else:
            # Throw error if not enough arguments specified
            for param_str in self._controller_params:
                if param_str in ("scheduler", "eval_batch_size"):
                    continue
                if getattr(self, param_str) is None:
                    raise ValueError(f"When controller is not specified, argument '{param_str}' must be specified.")


def setup_and_train(
    training_modules: TrainingModules,
    database,
    setup_params: SetupParams,
    store_all_better=False,
    store_best=True,
    store_every=0
):
    """
    :param: training_modules: see :func:`setup_training`
    :param: database: see :func:`train_model`
    :param: setup_params: see :func:`setup_training`
    :param: store_all_better: Save the state dict for each model doing better than a previous one
    :param: store_best: Save a checkpoint for the best model
    :param: store_every: Save a checkpoint for every certain epochs
    :return: See :func:`train_model`

    Shortcut for setup_training followed by train_model.

    .. Note::
        The training loop will capture KeyboardInterrupt exceptions to abort the experiment early.
        If you would like to gracefully kill training programmatically, see :func:`train_model` with callbacks argument.
    .. Note::
        Saves files in the current running directory; recommend you switch
        to a fresh directory with a descriptive name for your experiment.

    """
    # Set up objects for training.

    training_modules, controller, metric_tracker = setup_training(
        training_modules=training_modules, setup_params=setup_params
    )

    # Actually do the training
    return train_model(
        training_modules=training_modules,
        database=database,
        controller=controller,
        metric_tracker=metric_tracker,
        callbacks=None,
        batch_callbacks=None,
        store_all_better=store_all_better,
        store_best=store_best,
        store_every=store_every
    )


def setup_training(
    training_modules: TrainingModules,
    setup_params: SetupParams,
):
    """
    Prepares training_modules for training with experiment_params.

    :param: training_modules: Tuple of model, training loss, and evaluation losses
            (Can be built from graph using `graphs.assemble_training_modules`)
    :param: setup_params: parameters controlling how training is performed
            (See :class:`SetupParams`)

    Roughly:

    * sets devices for training modules
    * if no controller given:

       * instantiates and links optimizer to the learnable params on the model
       * instantiates and links scheduler to optimizer
       * builds a default controller with setup params
    * creates a MetricTracker for storing the training metrics

    :return: (optimizer,evaluator,controller,metrics,callbacks)
    """
    if isinstance(setup_params, dict):
        setup_params = SetupParams(**setup_params)

    model, loss, evaluator = training_modules

    controller = setup_params.controller
    if not isinstance(controller, Controller):
        print("Generating controller using values from setup params.")

        optimizer = setup_params.optimizer
        if not isinstance(optimizer, torch.optim.Optimizer) and callable(optimizer):
            print("Generating Optimizer from setup params, defaulting to parameters in model that require grad.")
            params = filter(lambda p: p.requires_grad, model.parameters())
            optimizer = optimizer(params, lr=setup_params.learning_rate)

        scheduler = setup_params.scheduler
        if not is_scheduler_like(scheduler) and callable(scheduler):
            print("Generating Scheduler from setup params")
            scheduler = scheduler(optimizer)

        controller_cls = Controller if controller is None else controller

        controller = controller_cls(
            optimizer=optimizer,
            scheduler=scheduler,
            batch_size=setup_params.batch_size,
            eval_batch_size=setup_params.eval_batch_size,
            max_epochs=setup_params.max_epochs,
            stopping_key=setup_params.stopping_key,
            fraction_train_eval=setup_params.fraction_train_eval,
        )

    optimizer = controller.optimizer
    model, evaluator, optimizer = serialization.set_devices(
        model, loss, evaluator, optimizer, setup_params.device or tools.device_fallback()
    )

    metrics = MetricTracker(evaluator.loss_names, stopping_key=controller.stopping_key)

    return training_modules, controller, metrics


def train_model(
    training_modules,
    database,
    controller,
    metric_tracker,
    callbacks,
    batch_callbacks,
    store_all_better=False,
    store_best=True,
    store_every=0,
    store_structure_file=True,
    store_metrics=True,
    quiet=False,
):
    """
    Performs training loop, allows keyboard interrupt. When done,
    reinstate the best model, make plots and metrics over time, and test the model.

    :param training_modules: tuple-like of model, loss, and evaluator
    :param database: Database
    :param controller: Controller
    :param metric_tracker: MetricTracker for storing model performance
    :param callbacks: callbacks to perform after every epoch.
    :param batch_callbacks: callbacks to perform after every batch
    :param store_best: Save a checkpoint for the best model
    :param store_all_better: Save the state dict for each model doing better than a previous one
    :param store_every: Save a checkpoint for every certain epochs
    :param store_structure_file: Save the structure file for this experiment
    :param store_metrics: Save the metric tracker for this experiment.
    :param quiet: If True, disable printing during training (still prints testing results).

    :return: metric_tracker

    .. Note::
        callbacks take the form of an iterable of callables
        and will be called with cb(epoch,new_best)

        * epoch indicates the epoch number
        * new_best indicates if the model is a new best model

    .. Note::
        batch_callbacks take the form of an  iterable of callables
        and will each be called with cb(batch_inputs, batch_model_outputs, batch_targets)

    .. Note::
        You may want to make your callbacks store other state, if so, an easy way is to make them a callable object.

    .. Note::
        callback state is not managed by ``hippynn``. If your wish to save or load callback state, you will
        have to manage that manually (possibly with a callback itself).

    """

    model, loss, evaluator = training_modules

    print("Beginning training.")
    if not quiet:
        if isinstance(model, (torch.nn.DataParallel, torch.nn.parallel.DistributedDataParallel)):
            print_model = model.module
            print("Distributed training with ", type(model))
        else:
            print_model = model
        try:
            print("Model:")
            print_model.print_structure()
        except Exception as eee:
            print(f"Model structure couldn't be printed. (Got {type(eee)}). Skipping.")
        print("Model Params:")
        tools.param_print(print_model)

    metric_tracker.quiet = quiet

    if store_structure_file:
        serialization.create_structure_file(training_modules, database, controller)

    try:
        training_loop(
            training_modules=training_modules,
            database=database,
            controller=controller,
            metric_tracker=metric_tracker,
            callbacks=callbacks,
            batch_callbacks=batch_callbacks,
            store_best=store_best,
            store_all_better=store_all_better,
            store_every=store_every,
            quiet=quiet,
        )

    except KeyboardInterrupt:
        print("******* TRAINING INTERRUPTED *******")
        print("Finishing up...")
    print("Training phase ended.")

    if store_metrics:
        with open("training_metrics.pkl", "wb") as pfile:
            pickle.dump(metric_tracker, pfile)

    best_model = metric_tracker.best_model
    if best_model:
        print("Reverting to best model found.")
        model.load_state_dict(best_model)

    print("Making plots over training time...")
    metric_tracker.plot_over_time()
    print("Testing model...")
    torch.cuda.empty_cache()
    test_model(
        database,
        evaluator,
        when="FinalTraining",
        batch_size=controller.eval_batch_size,
        metric_tracker=metric_tracker,
    )

    if store_metrics:
        with open("training_metrics.pkl", "wb") as pfile:
            pickle.dump(metric_tracker, pfile)
    print("Training complete.")
    return metric_tracker


def test_model(database, evaluator, batch_size, when, metric_tracker=None):
    """
    Tests the model on the database according to the model_evaluator metrics.
    If a plot_maker is attached to the model evaluator, it will make plots.
    The plots will go in a sub-folder specified by `when` the testing is taking place.
    The results are then printed.

    :param database: The database test the model on.
    :param evaluator: The evaluator containing model and evaluation losses to measure.
    :param when: A string to specify what plots are currently to be used.
    :param metric_tracker: (Optional) metric tracker to save metrics on. If not provided,
        a blank one will be constructed.

    :return: metric tracker
    """
    evaluator.model.eval()

    if metric_tracker is None:
        metric_tracker = MetricTracker(evaluator.loss_names, stopping_key=None)
    metric_tracker.quiet = False
    evaluation_data = collections.OrderedDict(
        (
            ("train", database.make_generator("train", "eval", batch_size)),
            ("valid", database.make_generator("valid", "eval", batch_size)),
            ("test", database.make_generator("test", "eval", batch_size)),
        )
    )
    evaluation_metrics = {k: evaluator.evaluate(gen, eval_type=k, when=when) for k, gen in evaluation_data.items()}
    metric_tracker.register_metrics(evaluation_metrics, when=when)
    metric_tracker.evaluation_print(evaluation_metrics)
    return metric_tracker


def training_loop(
    training_modules: TrainingModules,
    database,
    controller: Controller,
    metric_tracker: MetricTracker,
    callbacks,
    batch_callbacks,
    store_all_better,
    store_best,
    store_every,
    quiet,
):
    """
    Performs a high-level training loop.

    :param training_modules:  training modules from ``assemble_modules``
    :param database: database to train to
    :param controller: controller for early stopping and/or learning rate decay
    :param metric_tracker: the metric tracker
    :param callbacks: list of callbacks for each epoch
    :param batch_callbacks: list of callbacks for each batch
    :param store_best: Save a checkpoint for the best model
    :param store_all_better: Save the state dict for each model doing better than a previous one
    :param store_every: Save a checkpoint for every certain epochs
    :param quiet: whether to print information. Setting quiet to true won't prevent progress bars.

    :return: metrics -- the state of the experiment after training

    .. Note::
        Saves files in the current running directory; recommend switching
        to a fresh directory with a descriptive name for each experiment.

    Rough structure.

    Loop over Epochs, performing:

    - Loop over batches

        * Make predictions
        * Calculate Loss, perform backwards
        * Optimizer Step
        * Batch Callbacks
    - Perform validation and print results
    - Controller/Scheduler Step
    - Epoch callbacks
    - If new best, save the model state_dict and a checkpoint

    """

    if quiet:
        qprint = lambda *args, **kwargs: None
    else:
        qprint = print

    model, loss, evaluator = training_modules

    ### Trigger better error message if inputs or targets is not set!
    database.var_list
    # (If we remove this line, we'll still get an error but it will be more obscure.)
    ###
    n_inputs = len(database.inputs)
    n_targets = len(database.targets)

    qprint("At least {} epochs will be run".format(controller.max_epochs))

    epoch = metric_tracker.current_epoch
    device = evaluator.model_device
    step_function = get_step_function(controller.optimizer)
    optimizer = controller.optimizer

    continue_training = True  # Assume that nobody ran this function without wanting at least 1 epoch.

    while continue_training:

        qprint("_" * 50)
        qprint("Epoch {}:".format(epoch))
        tools.print_lr(optimizer)

        qprint(flush=True, end="")

        model.train()
        epoch_run_time = timeit.default_timer()
        train_generator = database.make_generator("train", "train", batch_size=controller.batch_size)

        for batch in tools.progress_bar(train_generator, desc="Training Batches", unit="batch"):

            batch = [item.to(device=device, non_blocking=True) for item in batch]
            batch_inputs = batch[:n_inputs]
            batch_targets = batch[-n_targets:]
            batch_targets = [x.requires_grad_(False) for x in batch_targets]

            batch_model_outputs = step_function(optimizer, model, loss, batch_inputs, batch_targets)

            if batch_callbacks:
                for cb in batch_callbacks:
                    cb(batch_inputs, batch_model_outputs, batch_targets)
            # Allow garbage collection of computed values.
            del batch_model_outputs

        elapsed_epoch_run_time = timeit.default_timer() - epoch_run_time
        qprint("Training time: ", round(elapsed_epoch_run_time, 2), "s")
        qprint("Validating...", flush=True)

        model.eval()

        evaluation_data = collections.OrderedDict(
            (
                (
                    "train",
                    database.make_generator(
                        "train", "eval", batch_size=controller.eval_batch_size, subsample=controller.fraction_train_eval
                    ),
                ),
                ("valid", database.make_generator("valid", "eval", controller.eval_batch_size)),
            )
        )
        evaluation_metrics = {k: evaluator.evaluate(gen, eval_type=k, when=epoch) for k, gen in evaluation_data.items()}

        better_metrics, better_model, stopping_metric = metric_tracker.register_metrics(evaluation_metrics, when=epoch)
        metric_tracker.evaluation_print_better(evaluation_metrics, better_metrics)

        # Check Learning Rate Schedule
        continue_training = controller.push_epoch(epoch, better_model, stopping_metric)

        elapsed_epoch_run_time = timeit.default_timer() - epoch_run_time
        metric_tracker.epoch_times.append(elapsed_epoch_run_time)
        qprint("Total epoch time: ", round(elapsed_epoch_run_time, 2), "s")

        if callbacks:
            for cb in callbacks:
                cb(epoch, better_model)

        if better_model:
            qprint("**** NEW BEST MODEL - Saving! ****")
            best_model = copy.deepcopy(model.state_dict())
            metric_tracker.best_model = best_model

            if store_all_better:
                # Save a copy of every network doing better
                # Note: epoch has already been incremented, so decrement in saving file.
                with open(f"better_model_epoch_{epoch}.pt", "wb") as pfile:
                    torch.save(best_model, pfile)

            if store_best:
                # Overwrite the "best model so far"
                with open("best_model.pt", "wb") as pfile:
                    torch.save(best_model, pfile)

                state = serialization.create_state(model, controller, metric_tracker)

                # Write the checkpoint
                with open("best_checkpoint.pt", "wb") as pfile:
                    torch.save(state, pfile)
            
        if store_every and epoch != 0 and (epoch % store_every) == 0:
            # Save a copy every "store_every" epoch
            with open(f"model_epoch_{epoch}.pt", "wb") as pfile:
                torch.save(model.state_dict(), pfile)
            
            state = serialization.create_state(model, controller, metric_tracker)

            # Write the checkpoint
            with open(f"checkpoint_epoch_{epoch}.pt", "wb") as pfile:
                torch.save(state, pfile)

        epoch += 1

    return metric_tracker



"""
Graph objects in hipppynn. This package allows definitions of interfaces for torch modules.
The interfaces allow dynamic reprogramming of control flow between pytorch modules.

Index Types (i.e. Molecule, Atom, Pair) track the physical basis for a tensor.
Automatic index type conversion can be performed between compatible types.

Furthermore, "parent expansions" allow for flexible signatures for node creation
which attempt to hide book-keeping from the end-user.
"""
from . import indextypes
from .indextypes import clear_index_cache, IdxType

from .nodes import base, inputs
from .nodes.base import find_unique_relative, find_relatives, get_connected_nodes

from .gops import get_subgraph, copy_subgraph, replace_node, compute_evaluation_order

from .nodes import networks, targets, physics, loss, excited

# Needed to populate the registry of index transformers.
# This has to happen before the indextypes package can work,
# however, we don't want the indextypes package to depend on actual
# implementations of nodes.
from . import indextransformers

from .graph import GraphModule

from .predictor import Predictor

__all__ = [
    "get_subgraph",
    "copy_subgraph",
    "replace_node",
    "compute_evaluation_order",
    "find_unique_relative",
    "find_relatives",
    "get_connected_nodes",
    "GraphModule",
    "Predictor",
    "IdxType",
]



"""
Base nodes for sublcassing.
"""
from ... import indextypes
from .algebra import _CombNode, AtLeast2D

from .node_functions import _BaseNode


class Node(_CombNode):
    pass


class SingleNode(Node):
    pass


class InputNode(SingleNode):
    _input_names = ()
    """Node for getting information for the database."""
    requires_grad = False
    input_type_str = "Input"

    def __init__(self, name=None, db_name=None):
        if db_name is None and name is None:
            raise TypeError("Input node requires name or db_name arguments.")
        if name is None and db_name is not None:
            name = self.input_type_str + "(db_name='{}')".format(db_name)
        super().__init__(name=name, parents=(), db_name=db_name, module=None)


class LossInputNode(InputNode):
    """Node for getting information from the model (predicted) or database (true) into the loss."""

    def __init__(self, name, origin_node, db_name):
        super().__init__(name, db_name)
        self.origin_node = origin_node

    @property
    def pred(self):
        raise TypeError("Node {} of type {} already in loss graph".format(self, type(self)))

    @property
    def true(self):
        raise TypeError("Node {} of type {} already in loss graph".format(self, type(self)))


class LossPredNode(LossInputNode):
    def __init__(self, name, origin_node, db_name):
        super().__init__(name, origin_node, db_name)
        self._index_state = getattr(origin_node, "_index_state", indextypes.IdxType.NotFound)


class LossTrueNode(LossInputNode):
    def __init__(self, name, origin_node, db_name):
        super().__init__(name, origin_node, db_name)
        self._index_state = indextypes.reduce_funcs.db_state_of(origin_node._index_state)
        self._main_output = AtLeast2D((self,))

    @property
    def main_output(self):
        return self._main_output


_BaseNode._LossPredNode = LossPredNode
_BaseNode._LossTrueNode = LossTrueNode



"""
Base node definition.
"""
from .. import _debprint

DEFAULT_WHY_DESC = "<purpose not specified>"


class _BaseNode:
    _input_names = NotImplemented
    _LossPredNode = None  # Will be set by this child class when it exists
    _LossTrueNode = None  # Same

    def __init__(self, name, parents, db_name=None, module=None):
        """

        :param name: name of node
        :param parents: parent nodes of the target
        :param db_name: name of target in database (if any)
        :param module: input pytorch module (if 'auto', attempt to create)


        """

        assert self not in parents, "Nodes cannot be their own parents."

        if not isinstance(name, str):
            raise TypeError("Node names must be strings. Instead got: {}".format(name))

        self.db_name = db_name
        self.origin_node = None  # Loss input nodes set this attribute to find references to the model graph
        self.parents = tuple(parents)
        self.name = name
        self._pred = None
        self._true = None
        self.children = tuple()
        for p in self.parents:
            p.children = p.children + (self,)

        # If specified, trigger automatic module generation
        if module == "auto":
            _debprint("Making auto module for", self.name)
            module = self.auto_module()
        # Otherwise, glue the module on
        if module is not None:
            self.torch_module = module

    def set_dbname(self, db_name):
        self.db_name = db_name
        if self._pred is not None:
            self._pred.db_name = db_name
        if self._true is not None:
            self._true.db_name = db_name

    @property
    def pred(self):
        if self._pred is None:
            self._pred = self._LossPredNode(self.name + "-pred", origin_node=self, db_name=self.db_name)
        return self._pred

    @property
    def true(self):
        if self._true is None:
            self._true = self._LossTrueNode(self.name + "-true", origin_node=self, db_name=self.db_name)
        return self._true

    def get_all_parents(self):
        return self.parents + tuple(pnode for parent in self.parents for pnode in parent.get_all_parents())

    def get_all_children(self):
        return self.children + tuple(ccnode for child in self.children for ccnode in child.get_all_children())

    # Functions that take either a node or a node set can be accessed as attributes.

    def get_all_connected(self):
        return get_connected_nodes({self})

    def find_unique_relative(self, constraint, why_desc=DEFAULT_WHY_DESC):
        return find_unique_relative(self, constraint, why_desc=why_desc)

    def find_relatives(self, constraint, why_desc=DEFAULT_WHY_DESC):
        return find_relatives(self, constraint, why_desc=why_desc)

    def swap_parent(self, old, new):
        if old not in self.parents:
            raise NodeOperationError(f"Node {old} is not a parent of node {self}")
        old.children = tuple(c for c in old.children if c is not self)
        new.children = *new.children, self
        self.parents = tuple(new if p is old else p for p in self.parents)

    def disconnect(self):
        """
        Remove this node from the graph, leaving it uncomputable.

        :return:
        """
        for p in self.parents:
            p.children = tuple(c for c in p.children if c is not self)
        self.parents = ()

    def disconnect_recursive(self):
        """
        Remove this node from the graph, then disconnect all children, recursively.

        Afterwards this node and all children thereof will no longer be computable.

        :return:
        """
        self.disconnect()
        for c in self.children:
            c.disconnect_recursive()

    def auto_module(self):
        raise NotImplementedError("Auto module not defined for node {} of type {}".format(self, type(self)))

    def __dir__(self):
        dir_ = super().__dir__()
        # need to protect against a case where input names are not specified.
        # otherwise dir() will raise an error. Debuggers hate that!
        if self._input_names is not NotImplemented:
            dir_ = dir_ + list(self._input_names)
        return dir_

    def __getattr__(self, item):
        if item in ("parents", "_input_names"):  # Guard against recursion
            raise AttributeError("Attribute {} not yet present".format(item))
        try:
            return self.parents[self._input_names.index(item)]
        except (AttributeError, ValueError) as ee:
            raise AttributeError("{} object has no attribute '{}'".format(self.__class__, item))

    def __repr__(self):
        try:
            name = self.name
        except AttributeError:
            name = "UNINITIALIZED"
        return "{}('{}')<{}>".format(self.__class__.__name__, name, hex(id(self)))

    # Overridden by MultiNode, LossInputNode
    @property
    def main_output(self):
        return self


class NodeOperationError(Exception):
    pass


class NodeNotFound(NodeOperationError):
    pass


def get_connected_nodes(node_set):
    """
    Recursively return nodes connected to the specified node_set.

    :param node_set: iterable collection of nodes (list, tuple, set,...)
    :return: set of nodes with some relationship to the input set.
    """
    search_from = set(node_set)
    search_found = set()
    # Very naive algorithm, but we don't anticipate large graphs.
    while len(search_from) != 0:
        for node in search_from.copy():
            search_found.add(node)
            search_from.remove(node)
            for node_relative in node.get_all_parents():
                if node_relative not in search_found:
                    search_from.add(node_relative)
            for node_relative in node.get_all_children():
                if node_relative not in search_found:
                    search_from.add(node_relative)
    return search_found


class NodeAmbiguityError(NodeOperationError):
    pass


def find_relatives(node_or_nodes, constraint_key, why_desc=DEFAULT_WHY_DESC):
    """

    :param node_or_nodes: a node or iterable of nodes to start the search.
    :param constraint_key: 1) callable to filter nodes by or
                        2) type spec to be used with `isinstance`.
    :param why_desc: If a node cannot be found satisfying the constraint, raise an error with this message.

    :return: set of nodes related to this node that obey a constraint
    """

    if isinstance(constraint_key, type):
        # We must bind the constraint key to a name within the lambda -- search_type
        constraint_key = lambda node, *, search_type=constraint_key: isinstance(node, search_type)
    elif callable(constraint_key):
        pass
    else:
        raise ValueError("constraint must be a type or callable filter function")

    if isinstance(node_or_nodes, _BaseNode):  # if we search from a node, wrap it as a collection
        node_or_nodes = [node_or_nodes]
        _debprint("Starting search from single node")

    candidates = {n for n in get_connected_nodes(node_or_nodes) if constraint_key(n)}

    for node in node_or_nodes:
        if constraint_key(node):
            candidates.add(node)

    if len(candidates) == 0:
        _debprint("Node not found, all relatives:")
        for n in get_connected_nodes(node_or_nodes):
            _debprint(n)
        raise NodeNotFound("({}) Missing: Could not automatically satisfying node in graph.".format(why_desc))

    return candidates


def find_unique_relative(node_or_nodes, constraint, why_desc=DEFAULT_WHY_DESC):
    """
    Look for a unique parent or child node type in the graph connected to the starting node.

    :param node_or_nodes: a node or iterable of nodes to start the search.
    :param constraint:
        1. callable to filter nodes by or
        2. type to be used with `isinstance`.
    :param why_desc: specification of error message

    :return: Node compatible with constraint.

    .. Note::
        If no node is found, a NodeNotFoundError is rasied.
        If more than one node is found, a NodeambiguityError is raised.
    """

    candidates = find_relatives(node_or_nodes, constraint_key=constraint, why_desc=why_desc)

    if len(candidates) > 1:
        raise NodeAmbiguityError("({}) Ambiguity: Multiple {} nodes found:{}".format(why_desc, constraint, candidates))

    result = candidates.pop()
    _debprint("Found node {} of type {}: {}".format(result, constraint.__name__, why_desc))
    return result



"""
Managing optimizer, scheduler, and end of training
"""
import inspect
import warnings

from torch.optim.lr_scheduler import ReduceLROnPlateau


class Controller:
    """
    Class for controlling the training dynamics.

    :param optimizer: pytorch optimizer (or compatible)
    :param scheduler: pytorch scheduler (or compatible)
        Pass None to use no schedulers. May pass a list of schedulers to use them all in sequence.
    :param scheduler_list: list of schedulers (may be empty).
    :param stopping_key: str of name for best metric.
    :param batch_size: batch size for training
    :param eval_batch_size: batch size during evaluation
    :param fraction_train_eval: the random fraction of the training set to use during evaluation.
    :param quiet: If true, suppress printing.

    :param max_epochs: maximum amount of epochs currently allowed based on training so far.

    .. Note::
       If a scheduler defines a ``set_controller`` method, this will be run on the Controller instance.
       This allows a scheduler to modify other aspects of training besides the conventional ones
       found in the optimizer. See ``RaiseBatchSizeOnPlateau`` in this module.

    """

    _state_vars = (
        "boredom",
        "current_epoch",
        "batch_size",
        "eval_batch_size",
        "_max_epochs",
        "fraction_train_eval",
        "stopping_key",
    )

    def __init__(
        self,
        optimizer,
        scheduler,
        batch_size,
        max_epochs,
        stopping_key,
        eval_batch_size=None,
        fraction_train_eval=0.1,
        quiet=False,
    ):

        self.optimizer = optimizer
        self.scheduler = scheduler

        self.stopping_key = stopping_key

        self.batch_size = batch_size
        self.eval_batch_size = eval_batch_size or batch_size
        if max_epochs is None:
            warnings.warn("Max epochs set to None. Model may train in an infinite loop.")

        self._max_epochs = max_epochs

        self.boredom = 0
        self.current_epoch = 0
        self.quiet = quiet
        self.fraction_train_eval = fraction_train_eval

        try:
            iter(scheduler)
        except TypeError:
            if scheduler is None:
                scheduler = []
            else:
                scheduler = [scheduler]

        for sch in scheduler:
            if hasattr(sch, "set_controller"):
                sch.set_controller(self)

        self.scheduler_list = scheduler

    def state_dict(self):
        state_dict = {k: getattr(self, k) for k in self._state_vars}
        state_dict["optimizer"] = self.optimizer.state_dict()
        state_dict["scheduler"] = [sch.state_dict() for sch in self.scheduler_list]
        return state_dict

    def load_state_dict(self, state_dict):

        for sch, sdict in zip(self.scheduler_list, state_dict["scheduler"]):
            sch.load_state_dict(sdict)

        self.optimizer.load_state_dict(state_dict["optimizer"])

        for k in self._state_vars:
            setattr(self, k, state_dict[k])

    @property
    def max_epochs(self):
        return self._max_epochs

    def push_epoch(self, epoch, better_model, metric):
        self.current_epoch += 1

        if better_model:
            self.boredom = 0
        else:
            self.boredom += 1

        for sch in self.scheduler_list:
            if accepts_metrics(sch):
                sch.step(metrics=metric)
            else:
                sch.step()

        if not self.quiet:
            print("Epochs since last best:", self.boredom)
            print("Current max epochs:", self.max_epochs)
        return self.current_epoch < self.max_epochs


class PatienceController(Controller):
    """
    A subclass of Controller that terminates if training has not improved for a given number
    of epochs.

    :ivar patience: How many epochs are allowed without improvement before termination.
    :ivar last_best: The eoch number of the last best epoch encountered.
    """

    _state_vars = Controller._state_vars + ("patience", "last_best")

    def __init__(self, *args, termination_patience, **kwargs):
        super().__init__(*args, **kwargs)
        self.patience = termination_patience
        self.last_best = 0

    def push_epoch(self, epoch, better_model, metric):
        if better_model:
            if self.boredom > 0 and not self.quiet:
                print("Patience for training restored.")
            self.boredom = 0
            self.last_best = epoch
        return super().push_epoch(epoch, better_model, metric)

    @property
    def max_epochs(self):
        return min(self.last_best + self.patience, self._max_epochs)


class RaiseBatchSizeOnPlateau:
    """
    Learning rate scheduler compatible with pytorch schedulers.

    This roughly implements the scheme outlined in the following paper:

    .. code-block:: none

        "Don't Decay the Learning Rate, Increase the Batch Size"
        Samuel L. Smith et al., 2018.
        arXiv:1711.00489
        Published as a conference paper at ICLR 2018.

    Until max_batch_size has been reached. After that,
    the learning rate is decayed.

    .. Note::
       To use this scheduler, build it, then link it to
       the container which governs the batch size using ``set_controller``.
       The default base hippynn Controller will do this automatically.
    """

    def __init__(
        self,
        optimizer,
        max_batch_size,
        factor=0.5,
        patience=10,
        threshold=0.0001,
        threshold_mode="rel",
        verbose=True,
        controller=None,
    ):

        if threshold_mode not in ("abs", "rel"):
            raise ValueError("Mode must be 'abs' or 'rel'")

        self.inner = ReduceLROnPlateau(
            optimizer,
            patience=patience,
            factor=factor,
            threshold=threshold,
            threshold_mode=threshold_mode,
            verbose=verbose,
        )
        self.controller = controller
        self.max_batch_size = max_batch_size
        self.best_metric = float("inf")
        self.boredom = 0
        self.last_epoch = 0

    def set_controller(self, box):
        self.controller = box

    def state_dict(self):
        return {
            "boredom": self.boredom,
            "best_metric": self.best_metric,
            "last_epoch": self.last_epoch,
            "inner": self.inner.state_dict(),
        }

    def load_state_dict(self, state_dict):
        self.boredom = state_dict["boredom"]
        self.best_metric = state_dict["boredom"]
        self.inner.load_state_dict(state_dict["inner"])
        self.last_epoch = state_dict["last_epoch"]

    def step(self, metrics):
        self.last_epoch += 1

        if self.controller is None:
            warnings.warn(
                "Batch size controller not specified. "
                "Defering to LR Scheduler and batch size will not be controlled."
            )

            return self.inner.step(metrics)

        if self.controller.batch_size >= self.max_batch_size:
            return self.inner.step(metrics)

        if self.inner.threshold_mode == "rel":
            better = self.best_metric * (1 - self.inner.threshold) > metrics
        elif self.inner.treshold_mode == "abs":
            better = self.best_metric - self.inner.threshold > metrics

        if better:
            self.boredom = 0
            self.best_metric = metrics
        else:
            self.boredom += 1

        if self.boredom > self.inner.patience:
            new_batch_size = int(self.controller.batch_size / self.inner.factor)
            new_batch_size = min(new_batch_size, self.max_batch_size)
            self.controller.batch_size = new_batch_size
            self.boredom = 0
            if self.inner.verbose:
                print("Raising batch size to", new_batch_size)
            if new_batch_size >= self.max_batch_size:
                self.inner.last_epoch = self.last_epoch - 1
                if self.inner.verbose:
                    print("Max batch size reached, Lowering learning rate from here.")

        return


def accepts_metrics(sch):
    """
    Return if a scheduler accepts the validation metric.

    :meta private:
    :param sch:
    :return: bool
    """
    signature = inspect.signature(sch.step)
    return "metrics" in signature.parameters


def is_scheduler_like(thing):
    """
    Test if an object can be used as a scheduler.
    :meta private:
    """
    # Pytorch doesn't have a unified inheritance class for these things.
    # So we call it a scheduler if it has a "step" attribute that itself is callable
    return callable(getattr(thing, "step", None))



"""
Enum for index states.
"""
import enum


# fmt: off
@enum.unique
class IdxType(enum.Enum):
    Scalar      = "Scalar"
    MolAtom     = "MolAtom"       # rectangular padded array of atom-like data
    Molecules   = "Molecules"
    Atoms       = "Atoms"
    Pair        = "Pair"
    MolAtomAtom = "MolAtomAtom"   # rectangular padded array of bond-like data
    QuadMol     = "QuadMol"
    QuadPack    = "QuadPack"      # packed 6-vec of quadrupole upper triangle
    NotFound    = "NOT FOUND"
    def __repr__(self): return f"<{self.__class__.__name__}.{self.name}>"
# fmt: on



"""
Functions for changing index states.
"""
import warnings

from . import _debprint
from .registry import _index_dispatch_table, elementwise_compare_rules, _db_index_states, _index_cache
from .type_def import IdxType


def dispatch_indexing(input_is, output_is):
    """
    Acquire the function for converting between two index states.

    :param input_is: input index state
    :param output_is: output index state
    :return: the function

    """
    _debprint("dispatching:", input_is, output_is)
    try:
        func = _index_dispatch_table[input_is, output_is]
    except KeyError as ke:
        raise ValueError("No index conversion ({})->({}) found.".format(input_is, output_is))
    return func


def index_type_coercion(node, output_index_state):
    """
    Attempt to convert a node to a given index state.

    .. Note::
       1. If given a MultiNode, this function operates on the node's main_output.
       2. If there is no IdxType on the node, this function is a no-op, so that if one has not been implemented,
          no error will be raised.
       3. If the index type already matches, then this function just returns the input node.
       4. If the index type doesn't match, an index transformer of the appropriate type will be looked for.
          If no conversion is found, a ValueError is raised.

    :param node:  the node to convert
    :param output_index_state: the index state to convert to.
    :return: a node in the requested index state
    """
    _debprint("Requesting transformation : {} -> {}".format(node.name, output_index_state))
    node = node.main_output
    try:
        needs_index = node._index_state != output_index_state
    except AttributeError as e:
        _debprint("\t Indexing information not found! Something is likely to fail.")
        needs_index = False

    if needs_index:
        prior_index_ref = (output_index_state, node)
        if (output_index_state, node) in _index_cache:
            # If this node is linked to a previous index state
            _debprint(f"\t Previous indexed version found for node {node} type {output_index_state}")
            outnode = _index_cache[prior_index_ref]
        else:
            _debprint(f"\t No index version found for node {node} type {output_index_state}")
            outnode = dispatch_indexing(node._index_state, output_index_state)(node)

    else:
        outnode = node
        _debprint("\t -> Autoindex unnecessary; index form correct.")

    return outnode


def soft_index_type_coercion(node, output_index_state):
    """
    Coerce if information is available.

    :param node:
    :param output_index_state:
    :return:
    """
    if output_index_state == IdxType.NotFound or node._index_state == IdxType.NotFound:
        _debprint(
            f"Soft request: Skipping requested conversion of {node} ({node._index_state})to {output_index_state}"
            " as information was not available."
        )
        return node
    else:
        return index_type_coercion(node, output_index_state)


def get_reduced_index_state(*nodes_to_reduce):
    """
    Find the index state for comparison between values in a loss function or plot.

    .. Note::
        This function is unlikely to be directed needed as a user.
        it's more likely you want to use :func:`elementwise_compare_reduce`.

    :param nodes_to_reduce:
    :return:
    """
    typeset = frozenset(getattr(n,"_index_state",IdxType.NotFound) for n in nodes_to_reduce)
    _debprint("Finding index comparison for for :", [n.name for n in nodes_to_reduce])
    try:
        coerced_type = elementwise_compare_rules[typeset]
        _debprint("Coerced type found: {}->{}".format(set(typeset), coerced_type))
    except KeyError as oke:
        _debprint("index rules:", elementwise_compare_rules)
        raise KeyError("index rule not found for {}".format(typeset)) from oke
    return coerced_type


def elementwise_compare_reduce(*nodes_to_reduce):
    """
    Return nodes converted to a mutually compatible index state with no padding.

    :param nodes_to_reduce: nodes to put in a comparable, reduced index state
    :return: node (if single argument) or tuple(nodes) (if multiple arguments)
    """
    if len(nodes_to_reduce) == 0:
        _debprint("Compare reduce on nothing, returning nothing")
        return tuple()
    coerced_type = get_reduced_index_state(*nodes_to_reduce)
    coerced_nodes = [index_type_coercion(node, coerced_type) for node in nodes_to_reduce]
    _debprint("Coerced:", [x.name for x in coerced_nodes])
    if len(coerced_nodes) == 1:
        return coerced_nodes[0]
    return coerced_nodes


def db_state_of(idxt):
    """
    Return the IdxType expected in the database for a given index type.

    .. Note::
        This function is unlikely to be directed needed as a user.
        it's more likely you want to use :func:`db_form`.
    """
    try:
        return _db_index_states[idxt]
    except KeyError:
        raise KeyError(f"No index state found for index type '{idxt}'")


def db_form(node):
    """
    Return a node converted to the index state of the database.
    """
    return index_type_coercion(node, db_state_of(node._index_state))



"""
Interface for creating LAMMPS MLIAP Unified models.
"""
import pickle
import warnings

import numpy as np
import torch

from lammps.mliap.mliap_unified_abc import MLIAPUnified

import hippynn
from hippynn.tools import device_fallback
from hippynn.graphs import find_relatives, find_unique_relative, get_subgraph, copy_subgraph, replace_node, IdxType, GraphModule
from hippynn.graphs.indextypes import index_type_coercion
from hippynn.graphs.gops import check_link_consistency
from hippynn.graphs.nodes.base import InputNode, SingleNode, MultiNode, AutoNoKw, ExpandParents
from hippynn.graphs.nodes.tags import Encoder, PairIndexer
from hippynn.graphs.nodes.indexers import PaddingIndexer
from hippynn.graphs.nodes.physics import GradientNode, VecMag
from hippynn.graphs.nodes.inputs import SpeciesNode
from hippynn.graphs.nodes.pairs import PairFilter


class MLIAPInterface(MLIAPUnified):
    """
    Class for creating ML-IAP Unified model based on hippynn graphs.
    """

    def __init__(self, energy_node, element_types, ndescriptors=1, model_device=torch.device("cpu"), compute_dtype=torch.float32):
        """
        :param energy_node: Node for energy
        :param element_types: list of atomic symbols corresponding to element types
        :param ndescriptors: the number of descriptors to report to LAMMPS
        :param model_device: the device to send torch data to (cpu or cuda)
        """
        super().__init__()
        self.element_types = element_types
        self.ndescriptors = ndescriptors
        self.model_device = model_device

        # Build the calculator
        self.rcutfac, self.species_set, self.graph = setup_LAMMPS_graph(energy_node)
        self.nparams = sum(p.nelement() for p in self.graph.parameters())
        self.compute_dtype = compute_dtype
        self.graph.to(compute_dtype)

    def compute_gradients(self, data):
        pass

    def compute_descriptors(self, data):
        pass

    def as_tensor(self, array):
        return torch.as_tensor(array, device=self.model_device)

    def empty_tensor(self,dimentions):
        return torch.empty(dimentions,device=self.model_device)

    def compute_forces(self, data):
        """
        :param data: MLIAPData object (provided internally by lammps)
        :return None
        This function writes results to the input `data`.
        """
        nlocal = self.as_tensor(data.nlistatoms)
        if nlocal.item() > 0:
            #If there are no local atoms, do nothing
            elems = self.as_tensor(data.elems).type(torch.int64).reshape(1, data.ntotal)
            z_vals = self.species_set[elems + 1]
            npairs = data.npairs
            if npairs > 0:
                pair_i = self.as_tensor(data.pair_i).type(torch.int64)
                pair_j = self.as_tensor(data.pair_j).type(torch.int64)
                rij = self.as_tensor(data.rij).type(self.compute_dtype)
            else:
                pair_i = self.empty_tensor(0).type(torch.int64)
                pair_j = self.empty_tensor(0).type(torch.int64)
                rij = self.empty_tensor([0,3]).type(self.compute_dtype)
    
            # note your sign for rij might need to be +1 or -1, depending on how your implementation works
            inputs = [z_vals, pair_i, pair_j, -rij, nlocal]
            atom_energy, total_energy, fij = self.graph(*inputs)
    
            # Test if we are using lammps-kokkos or not. Is there a more clear way to do that?
            if isinstance(data.elems, np.ndarray):
                return_device = "cpu"
            else:
                # Hope that kokkos device and pytorch device are the same (default cuda)
                return_device = elems.device
    
            atom_energy = atom_energy.squeeze(1).detach().to(return_device)
            total_energy = total_energy.detach().to(return_device)

            f = self.as_tensor(data.f)
            fij = fij.type(f.dtype).detach().to(return_device)
    
            if return_device == "cpu":
                fij = fij.numpy()
                data.eatoms = atom_energy.numpy().astype(np.double)
            else:
                eatoms = torch.as_tensor(data.eatoms, device=return_device)
                eatoms.copy_(atom_energy)
            if npairs > 0:
                data.update_pair_forces(fij)
            data.energy = total_energy.item()

    def __getstate__(self):
        self.species_set = self.species_set.to(torch.device("cpu"))
        self.graph.to(torch.device("cpu"))
        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__.update(state)
        try:
            torch.ones(0).to(self.model_device)
        except RuntimeError:
            fallback = device_fallback()
            warnings.warn(f"Model device ({self.model_device}) not found, falling back to f{fallback}")
            self.model_device = fallback

        self.species_set = self.species_set.to(self.model_device)
        self.graph.to(self.model_device)


def setup_LAMMPS_graph(energy):
    """

    :param energy: energy node for lammps interface
    :return: graph for computing from lammps MLIAP unified inputs.
    """
    required_nodes = [energy]

    why = "Generating LAMMPS Calculator interface"
    subgraph = get_subgraph(required_nodes)

    search_fn = lambda targ, sg: lambda n: n in sg and isinstance(n, targ)
    pair_indexers = find_relatives(required_nodes, search_fn(PairIndexer, subgraph), why_desc=why)

    new_required, new_subgraph = copy_subgraph(required_nodes, assume_inputed=pair_indexers, tag="LAMMPS")
    pair_indexers = find_relatives(new_required, search_fn(PairIndexer, new_subgraph), why_desc=why)

    species = find_unique_relative(new_required, search_fn(SpeciesNode, new_subgraph), why_desc=why)

    encoder = find_unique_relative(species, search_fn(Encoder, new_subgraph), why_desc=why)
    padding_indexer = find_unique_relative(species, search_fn(PaddingIndexer, new_subgraph), why_desc=why)
    inv_real_atoms = padding_indexer.inv_real_atoms

    species_set = torch.as_tensor(encoder.species_set).to(torch.int64)
    min_radius = max(p.dist_hard_max for p in pair_indexers)

    ###############################################################
    # Set up graph to accept external pair indices and shifts

    in_pair_first = InputNode("(LAMMPS)pair_first")
    in_pair_first._index_state = hippynn.graphs.IdxType.Pair
    in_pair_second = InputNode("(LAMMPS)pair_second")
    in_pair_second._index_state = hippynn.graphs.IdxType.Pair
    in_pair_coord = InputNode("(LAMMPS)pair_coord")
    in_pair_coord._index_state = hippynn.graphs.IdxType.Pair
    in_nlocal = InputNode("(LAMMPS)nlocal")
    in_nlocal._index_state = hippynn.graphs.IdxType.Scalar
    pair_dist = VecMag("(LAMMPS)pair_dist", in_pair_coord)
    mapped_pair_first = ReIndexAtomNode("pair_first_internal", (in_pair_first, inv_real_atoms))
    mapped_pair_second = ReIndexAtomNode("pair_second_internal", (in_pair_second, inv_real_atoms))

    new_inputs = [species, in_pair_first, in_pair_second, in_pair_coord, in_nlocal]

    # Construct Filters and replace the existing pair indexers with the
    # corresponding new (filtered) node that accepts external pairs of atoms
    for pi in pair_indexers:
        if pi.dist_hard_max == min_radius:
            replace_node(pi.pair_first, mapped_pair_first, disconnect_old=False)
            replace_node(pi.pair_second, mapped_pair_second, disconnect_old=False)
            replace_node(pi.pair_coord, in_pair_coord, disconnect_old=False)
            replace_node(pi.pair_dist, pair_dist, disconnect_old=False)
            pi.disconnect()
        else:
            mapped_node = PairFilter(
                "DistanceFilter-LAMMPS",
                (pair_dist, in_pair_first, in_pair_second, in_pair_coord),
                dist_hard_max=pi.dist_hard_max,
            )
            replace_node(pi.pair_first, mapped_node.pair_first, disconnect_old=False)
            replace_node(pi.pair_second, mapped_node.pair_second, disconnect_old=False)
            replace_node(pi.pair_coord, mapped_node.pair_coord, disconnect_old=False)
            replace_node(pi.pair_dist, mapped_node.pair_dist, disconnect_old=False)
            pi.disconnect()

    energy, *new_required = new_required
    try:
        atom_energies = energy.atom_energies
    except AttributeError:
        atom_energies = energy

    try:
        atom_energies = index_type_coercion(atom_energies, IdxType.Atoms)
    except ValueError:
        raise RuntimeError(
            "Could not build LAMMPS interface. Pass an object with index type IdxType.Atoms or "
            "an object with an `atom_energies` attribute."
        )

    local_atom_energy = LocalAtomEnergyNode("(LAMMPS)local_atom_energy", (atom_energies, in_nlocal))
    grad_rij = GradientNode("(LAMMPS)grad_rij", (local_atom_energy.total_local_energy, in_pair_coord), -1)

    implemented_nodes = local_atom_energy.local_atom_energies, local_atom_energy.total_local_energy, grad_rij

    check_link_consistency((*new_inputs, *implemented_nodes))
    mod = GraphModule(new_inputs, implemented_nodes)
    mod.eval()

    return min_radius / 2, species_set, mod


class ReIndexAtomMod(torch.nn.Module):
    def forward(self, raw_atom_index_array, inverse_real_atoms):
        return inverse_real_atoms[raw_atom_index_array]


class ReIndexAtomNode(AutoNoKw, SingleNode):
    _input_names = "raw_atom_index_array", "inverse_real_atoms"
    _main_output = "total_local_energy"
    _auto_module_class = ReIndexAtomMod

    def __init__(self, name, parents, module="auto", **kwargs):
        self._index_state = parents[0]._index_state
        super().__init__(name, parents, module=module, **kwargs)


class LocalAtomsEnergy(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, all_atom_energies, nlocal):
        local_atom_energies = all_atom_energies[:nlocal]
        total_local_energy = torch.sum(local_atom_energies)
        return local_atom_energies, total_local_energy


class LocalAtomEnergyNode(AutoNoKw, ExpandParents, MultiNode):
    _input_names = "all_atom_energies", "nlocal"
    _output_names = "local_atom_energies", "total_local_energy"
    _main_output = "total_local_energy"
    _output_index_states = None, IdxType.Scalar
    _auto_module_class = LocalAtomsEnergy

    _parent_expander.assertlen(2)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, IdxType.Scalar)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)



"""
Wraps non-pytorch implementations for use with pytorch autograd.
"""
import torch.autograd


def wrap_envops(envsum_impl, sensesum_impl, featsum_impl):
    """
    :param envsum_impl:  non-autograd implementation of envsum
    :param sensesum_impl:  non-autograd implementation of sensesum
    :param featsum_impl:  non-autograd implementtation of featsum
    :return:
    """

    class AGEnvsum(torch.autograd.Function):
        @staticmethod
        def forward(ctx, sense, feat, pfirst, psecond):
            ctx.save_for_backward(sense, feat, pfirst, psecond)
            if pfirst.shape[0] == 0:
                n_pair, n_nu = sense.shape
                n_atom, n_feat = feat.shape
                if n_pair!=0 or psecond.shape[0]!=0:
                    raise ValueError("Inconsistent shapes for envsum.")
                return torch.zeros((n_atom,n_nu,n_feat),dtype=feat.dtype,device=feat.device)
            env = envsum_impl(sense, feat, pfirst, psecond)
            return env

        @staticmethod
        def backward(ctx, grad_output):
            (
                sense,
                feat,
                pfirst,
                psecond,
            ) = ctx.saved_tensors
            need_gradsense, need_gradfeat, *_ = ctx.needs_input_grad

            need_gradsense = need_gradsense or None
            need_gradfeat = need_gradfeat or None
            grad_sense = need_gradsense and sensesum(grad_output, feat, pfirst, psecond)
            grad_feat = need_gradfeat and featsum(grad_output, sense, pfirst, psecond)

            return grad_sense, grad_feat, None, None

    class AGSensesum(torch.autograd.Function):
        @staticmethod
        def forward(ctx, env, feat, pfirst, psecond):
            ctx.save_for_backward(env, feat, pfirst, psecond)
            if pfirst.shape[0] == 0:
                n_atom0, n_nu, n_feat0 = env.shape
                n_atom1, n_feat1 = feat.shape
                if psecond.shape[0] !=0 or n_atom0!=n_atom1 or n_feat0 != n_feat1:
                    raise ValueError("Inconsistent shapes for sensesum")
                return torch.zeros((0,n_nu),dtype=feat.dtype,device=feat.device)
            sense = sensesum_impl(env, feat, pfirst, psecond)
            return sense

        @staticmethod
        def backward(ctx, grad_output):
            env, feat, pfirst, psecond = ctx.saved_tensors
            needs_gradenv, needs_gradfeat, *_ = ctx.needs_input_grad
            needs_gradenv = needs_gradenv or None
            needs_gradfeat = needs_gradfeat or None
            gradenv = needs_gradenv and envsum(grad_output, feat, pfirst, psecond)
            gradfeat = needs_gradfeat and featsum(env, grad_output, pfirst, psecond)
            return gradenv, gradfeat, None, None

    class AGFeatsum(torch.autograd.Function):
        @staticmethod
        def forward(ctx, env, sense, pfirst, psecond):
            ctx.save_for_backward(env, sense, pfirst, psecond)
            if pfirst.shape[0] == 0:
                n_atom, n_nu0, n_feat = env.shape
                n_pair, n_nu1 = sense.shape
                if psecond.shape[0] !=0 or n_nu0!=n_nu1:
                    raise ValueError("Inconsistent shapes for featsum")
                return torch.zeros((n_atom,n_feat),dtype=env.dtype,device=env.device)
            feat = featsum_impl(env, sense, pfirst, psecond)
            return feat

        @staticmethod
        def backward(ctx, grad_output):
            env, sense, pfirst, psecond = ctx.saved_tensors
            needs_gradenv, needs_gradsense, *_ = ctx.needs_input_grad
            needs_gradenv = needs_gradenv or None
            needs_gradsense = needs_gradsense or None
            gradgenv = needs_gradenv and envsum(sense, grad_output, pfirst, psecond)
            gradsense = needs_gradsense and sensesum(env, grad_output, pfirst, psecond)
            return gradgenv, gradsense, None, None

    envsum = AGEnvsum.apply
    sensesum = AGSensesum.apply
    featsum = AGFeatsum.apply

    return envsum, sensesum, featsum



"""
Layers for getting L2 parameters
"""
import torch


class LPReg(torch.nn.Module):
    def __init__(self, network, p=2):
        super().__init__()
        self.network = network
        self.p = p

    def forward(self, *args):
        # args is ignored; the output is not dependent on the output of the network.
        params = list(self.network.regularization_params())
        if len(params) == 0:
            raise ValueError("L^p regularization has no parameters.")
        return sum(torch.sum(torch.abs(torch.pow(p, self.p))) for p in params)



"""
Definitions of nodes for graph computation.
"""
import warnings

from ... import settings

if settings.DEBUG_NODE_CREATION:
    warnings.warn("Printing automatic node creation info! Output will be verbose.")


def _debprint(*args, **kwargs):
    if settings.DEBUG_NODE_CREATION:
        print("AutoNode:", *args, **kwargs)
    else:
        pass


from . import base, inputs, indexers, networks, targets, physics, loss



"""
Tags for nodes to inherit from, allowing tagging of different kinds
of information.
This file should not depend on any actual nodes.
"""
from ..indextypes import IdxType


class Encoder:
    _output_names = "encoding", "nonblank"
    species_set = NotImplemented


class AtomIndexer:
    pass


class Network:
    pass


class PairIndexer:
    _output_names = "pair_dist", "pair_first", "pair_second", "pair_coord"
    _output_index_states = (IdxType.Pair,) * len(_output_names)


class PairCache:
    pass


class HAtomRegressor:
    pass


class Charges:
    pass


class Positions:
    pass


class Species:
    pass


class Energies:
    pass



"""
A base node that provides several output tensors.
"""
from ....layers.algebra import Idx
from .base import SingleNode, Node
from .. import _debprint
from ...indextypes import IdxType


class IndexNode(SingleNode):
    _input_names = ("parent",)

    def __init__(self, name, parents, index, index_state=None):
        if len(parents) != 1:
            raise TypeError("Index node takes exactly one parent.")
        par = parents[0]
        iname = par._output_names[index] if hasattr(par, "_output_names") else "<{index}>".format(index=index)
        repr_info = {"parent_name": par.name, "index": iname}
        module = Idx(index, repr_info=repr_info)
        self.index = index
        self._index_state = IdxType.NotFound if index_state is None else index_state
        super().__init__(name, parents, module=module)


class MultiNode(Node):  # Multinode
    _output_names = NotImplemented
    _output_index_states = NotImplemented  # optional?
    _main_output = NotImplemented

    def __init__(self, name, parents, module="auto", *args, db_name=None, **kwargs):

        super().__init__(name, parents, *args, module=module, **kwargs)

        self.children = tuple(
            IndexNode(name + "." + cn, (self,), index=i, index_state=cidx)
            for i, (cn, cidx) in enumerate(zip(self._output_names, self._output_index_states))
        )
        self.main_output.db_name = db_name

    def set_dbname(self, db_name):
        self.main_output.set_dbname(db_name)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Enforce _child_index_states has same length as _output_names
        if cls._output_index_states is not NotImplemented:
            if len(cls._output_index_states) != len(cls._output_names):
                raise AssertionError(
                    "Lengths of _child_index_states {} doesn't match lengths of ouput_names {}".format(
                        cls._output_index_states, cls._output_names
                    )
                )

        # Enforce no name conflict between input names and output names
        if cls._input_names is not NotImplemented:
            try:
                assert all(o not in cls._input_names for o in cls._output_names)
            except AssertionError as ae:
                raise ValueError(
                    "Multi-node output names {} conflict with input names {}".format(
                        cls._output_names, cls._input_names
                    )
                ) from ae

    def __dir__(self):
        dir_ = super().__dir__()
        if self._output_names is not NotImplemented:
            dir_ = dir_ + list(self._output_names)
        return dir_

    def __getattr__(self, item):
        if item in ("children", "_output_names"):  # Guard against recursion
            raise AttributeError("Attribute {} not yet present.".format(item))
        try:
            return super().__getattr__(item)  # Defer to BaseNode first
        except AttributeError:
            pass
        try:
            return self.children[self._output_names.index(item)]
        except (AttributeError, ValueError):
            raise AttributeError("{} object has no attribute '{}'".format(self.__class__, item))

    @property
    def main_output(self):
        if self._main_output is NotImplemented:
            return super().main_output
        return getattr(self, self._main_output)



"""

The hippynn python package.

"""

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

# Configurational settings
from ._settings_setup import settings


# Pytorch modules
from . import layers
from . import networks

# Graph abstractions
from . import graphs

# Database loading
from . import databases

# Training/testing routines
from . import experiment
from .experiment import setup_and_train

# Custom Kernels
from . import custom_kernels

from . import pretraining

from . import tools



"""
Read Databases in the ANI H5 format.

Note: You will need `pyanitools.py` to be importable to import this module.
"""

import pyanitools

import os

import numpy as np
import torch

from ..tools import progress_bar, np_of_torchdefaultdtype
from ..tools import pad_np_array_to_length_with_zeros as pad_atoms

from . import Database
from .restarter import Restartable

from ase.data import atomic_numbers

numpy_map_elements = np.vectorize(atomic_numbers.__getitem__)


class PyAniMethods:
    _IGNORE_KEYS = ("path", "Jnames")

    # Note: assumes that all data files have 'coordinates'.
    def extract_full_file(self, file, species_key="species"):
        n_atoms_max = 0
        batches = []
        x = pyanitools.anidataloader(file)

        for c in progress_bar(x, desc="Data Groups", unit="group", total=x.group_size()):
            batch_dict = {}
            if species_key not in c:
                raise ValueError(f"Species key '{species_key}' not found' in file {file}!\n"
                                 f"\tFound keys: {set(c.keys())}")
            for k, v in c.items():
                # Filter things we don't need
                if k in self._IGNORE_KEYS:
                    continue

                # Convert to numpy
                v = v if np.isscalar(v) else np.asarray(v)

                # Special logic for species
                if k == species_key:
                    # Groups have the same species, broad-cast out the batch axis
                    v = np.expand_dims(v, 0)
                    # If given as strings, map to atomic elements
                    if (not isinstance(v.dtype, type)) and issubclass(v.dtype.type, np.str_):
                        v = numpy_map_elements(v)

                    n_atoms_max = max(n_atoms_max, v.shape[1])

                batch_dict[k] = v

            batches.append(batch_dict)

        return batches, n_atoms_max

    def determine_key_structure(self, batch_list, species_key="species"):
        """Determine what arrays to pad"""
        batch = batch_list[0]
        n_atoms = batch[species_key].shape[1]

        # algorithm will fail if n_atoms is confusable with other dimensions,
        # so just recurse to the next batch if this is detected.
        # The number 7 is relatively arbitrary, but if we use, for example, 3 then
        # the algorithm would treat the coordinates as needing padding.
        if n_atoms < 7:
            try:
                return self.determine_key_structure(batch_list[1:], species_key=species_key)
            except RecursionError as re:
                msg = "Automatic detection of arrays is only compatible with datasets of at least 6 atoms -- this is not supported."
                raise ValueError(msg) from re

        # dict of which axes need to be padded;
        # pad if the array size is equal to the number of atoms along a given axis
        padding_scheme = {k: [] for k in batch.keys()}
        bsize = 0
        bkey = None
        for k, v in batch.items():
            for i, l in enumerate(v.shape):
                if i == 0:
                    continue  # Don't pad the batch index
                if l == n_atoms:
                    padding_scheme[k].append(i)
                    # Use the largest 0th-axis shape that has an atom index
                    # as the indicator key for the batch size
                    this_bsize = v.shape[0]
                    if this_bsize > bsize:
                        bsize = this_bsize
                        bkey = k

        padding_scheme['sys_number'] = []
        return padding_scheme, bkey

    def process_batches(self, batches, n_atoms_max, species_key="species"):

        # Get padding info and batch size key
        padding_scheme, size_key = self.determine_key_structure(batches, species_key=species_key)

        # Pad the arrays
        padded_batches = []
        for i, b in enumerate(progress_bar(batches, desc="Processing Batches", unit="batch")):
            pb = {}
            b['sys_number'] = np.asarray([i], dtype=np.int64)
            for k, v in b.items():
                bsize = len(b[size_key])
                # Expand species array to fit batch size
                if k == species_key:

                    v = np.repeat(v, bsize, axis=0)

                # Perform padding as needed
                for axis in padding_scheme[k]:
                    v = pad_atoms(v, n_atoms_max, axis=axis)
                if 0 not in padding_scheme[k] and v.shape[0] == 1:
                    v = np.broadcast_to(v, (bsize, *v.shape[1:]))

                pb[k] = v

            padded_batches.append(pb)

        arr_dict = {}

        for k in b.keys():
            try:
                arr_dict[k] = np.concatenate([pb[k] for pb in padded_batches])
            except ValueError as ve:
                print("Error occured:",ve)
                print("Skipping key:",k)
                continue

        return arr_dict

    def filter_arrays(self, arr_dict, allow_unfound=False, quiet=False):
        if not quiet:
            print("Arrays found: ", list(arr_dict.keys()))

        floatX = np_of_torchdefaultdtype()
        for k, v in arr_dict.copy().items():
            if not allow_unfound:
                if k not in self.inputs and k not in self.targets:
                    del arr_dict[k]
                    continue
            if v.dtype == "float64":
                arr_dict[k] = v.astype(floatX)

        if not quiet:
            print("Data types:")
            print({k: v.dtype for k, v in arr_dict.items()})

        return arr_dict


class PyAniFileDB(Database, PyAniMethods, Restartable):
    def __init__(self, file, inputs, targets, *args, allow_unfound=False,species_key="species", quiet=False, **kwargs):

        self.file = file
        self.inputs = inputs
        self.targets = targets
        self.species_key = species_key

        arr_dict = self.load_arrays(quiet=quiet, allow_unfound=allow_unfound)

        super().__init__(arr_dict, inputs, targets, *args, **kwargs, quiet=quiet, allow_unfound=allow_unfound)
        self.restarter = self.make_restarter(
            file, inputs, targets, *args, **kwargs, quiet=quiet, allow_unfound=allow_unfound,
            species_key=species_key,
        )

    def load_arrays(self, allow_unfound=False, quiet=False):
        if not quiet:
            print("Loading arrays from", self.file)
        batches, n_atoms_max = self.extract_full_file(self.file,species_key=self.species_key)
        arr_dict = self.process_batches(batches, n_atoms_max,species_key=self.species_key)
        arr_dict = self.filter_arrays(arr_dict, quiet=quiet, allow_unfound=allow_unfound)
        return arr_dict


class PyAniDirectoryDB(Database, PyAniMethods, Restartable):
    def __init__(self, directory, inputs, targets, *args, files=None, allow_unfound=False,species_key="species",
                 quiet=False,**kwargs):

        self.directory = directory
        self.files = files
        self.inputs = inputs
        self.targets = targets
        self.species_key = species_key
        arr_dict = self.load_arrays(allow_unfound=allow_unfound,quiet=quiet)

        super().__init__(arr_dict, inputs, targets, *args, **kwargs, quiet=quiet, allow_unfound=allow_unfound)
        self.restarter = self.make_restarter(directory, inputs, targets, *args, files=files, quiet=quiet,
                                             species_key=species_key, **kwargs)

    def load_arrays(self, allow_unfound=False, quiet=False):

        if self.files:
            files = [os.path.join(self.directory, f) for f in self.files]
        else:
            files = [os.path.join(self.directory, f) for f in os.listdir(self.directory) if f.endswith(".h5")]
        files.sort()

        if len(files) == 0:
            raise FileNotFoundError(f"No '.h5' files found in directory {self.directory}")

        if not quiet:
            print("Gathering data from files:\n\t", end="")
            print(*files, sep="\n\t")

        file_batches = []
        for f in progress_bar(files, desc="Data Files", unit="file"):
            file_batches.append(self.extract_full_file(f,species_key=self.species_key))

        data, max_atoms_list = zip(*file_batches)

        n_atoms_max = max(max_atoms_list)
        batches = [item for fb in data for item in fb]

        arr_dict = self.process_batches(batches, n_atoms_max,species_key=self.species_key)
        arr_dict = self.filter_arrays(arr_dict, quiet=quiet, allow_unfound=allow_unfound)
        return arr_dict



"""
Dataset stored as NPY files in directory or
as NPZ dictionary.
"""
import os

import numpy as np
import torch

from ..tools import np_of_torchdefaultdtype
from .database import Database
from .restarter import Restartable


class DirectoryDatabase(Database, Restartable):
    """
    Database stored as NPY files in a diectory.

    :param directory: directory path where the files are stored
    :param name: prefix for the arrays.

    This function loads arrays of the format f"{name}{db_name}.npy" for each variable db_name in inputs and targets.

    Other arguments: See ``Database``.

    .. Note::
       This database loader does not support the ``allow_unfound`` setting in the base ``Database``. The
       variables to load must be set explicitly in the inputs and targets.
    """

    def __init__(self, directory, name, inputs, targets, *args, quiet=False, allow_unfound=False, **kwargs):
        if allow_unfound:
            raise ValueError("DirectoryDatabase class does not support allow_unfound argument.")

        arr_dict = self.load_arrays(directory, name, inputs, targets, quiet=quiet)
        super().__init__(arr_dict, inputs, targets, *args, **kwargs, quiet=quiet)

        self.restarter = self.make_restarter(
            directory,
            name,
            inputs,
            targets,
            *args,
            **kwargs,
            quiet=quiet,
        )

    def get_file_dict(self, directory, prefix):
        try:
            file_list = os.listdir(directory)
        except FileNotFoundError as fee:
            raise FileNotFoundError(
                "ERROR: Couldn't find directory {} containing files."
                'A solution is to explicitly specify "path" in database_params '.format(directory)
            ) from fee

        data_labels = {
            file[len(prefix) : -4]: file for file in file_list if file.startswith(prefix) and file.endswith(".npy")
        }

        # Make sure we actually found some files
        if not data_labels:
            raise FileNotFoundError(
                "No files found at {} .".format(directory) + "for database prefix {}".format(prefix)
            )
        return data_labels

    def load_arrays(self, directory, name, inputs, targets, quiet=False, allow_unfound=False):

        var_list = inputs + targets
        # Make sure the path actually exists

        try:
            # Backward compatibility.
            data_labels = self.get_file_dict(directory, prefix="data-" + name)
        except FileNotFoundError:
            data_labels = self.get_file_dict(directory, prefix=name)

        if not quiet:
            print("Arrays found: ", data_labels)

        # Load files
        arr_dict = {
            label: np.load(os.path.join(directory, file))
            for label, file in data_labels.items()
            if allow_unfound or (label in var_list)
        }

        # Put float64 data in pytorch default dtype
        floatX = np_of_torchdefaultdtype()
        for k, v in arr_dict.items():
            if v.dtype == "float64":
                arr_dict[k] = v.astype(floatX)

        if not quiet:
            print("Data types:")
            print({k: v.dtype for k, v in arr_dict.items()})

        return arr_dict


class NPZDatabase(Database, Restartable):
    def __init__(self, file, inputs, targets, *args, allow_unfound=False, quiet=False, **kwargs):
        arr_dict = self.load_arrays(file, inputs, targets, quiet=quiet, allow_unfound=allow_unfound)
        super().__init__(arr_dict, inputs, targets, *args, **kwargs, quiet=quiet,allow_unfound=allow_unfound)
        self.restarter = self.make_restarter(
            file, inputs, targets, *args, **kwargs, quiet=quiet, allow_unfound=allow_unfound
        )

    def load_arrays(self, file, inputs, targets, allow_unfound=False, quiet=False):

        arr_dict = np.load(file)
        # Make sure the path actually exists
        if not quiet:
            print("Arrays found: ", list(arr_dict.keys()))

        # Load files
        if not allow_unfound:
            var_list = inputs + targets
            arr_dict = {k: v for k, v in arr_dict.items() if k in var_list}
        else:
            arr_dict = {k: v for k,v, in arr_dict.items()}

        # Put float64 data in pytorch default dtype
        floatX = np_of_torchdefaultdtype()
        for k, v in arr_dict.items():
            if v.dtype == "float64":
                arr_dict[k] = v.astype(floatX)

        if not quiet:
            print("Data types:")
            print({k: v.dtype for k, v in arr_dict.items()})

        return arr_dict



"""
Nodes for usage with hippynn node system.
"""
import torch

from .seqm_modules import Scale
from .seqm_modules import SEQM_Energy, SEQM_All
from .seqm_modules import SEQM_MolMask, AtomMask, SEQM_OrbitalMask
from .seqm_modules import SEQM_MaskOnMol, SEQM_MaskOnMolAtom, SEQM_MaskOnMolOrbital, SEQM_MaskOnMolOrbitalAtom
from hippynn.graphs.nodes.base import MultiNode, AutoKw, find_unique_relative, ExpandParents, SingleNode
from hippynn.graphs.indextypes import IdxType
from hippynn.graphs.nodes.networks import Network
from hippynn.graphs.nodes.targets import HChargeNode, HBondNode
from hippynn.graphs.nodes.inputs import PositionsNode, SpeciesNode


class ScaleNode(AutoKw, SingleNode):
    _input_names = "notconverged"
    _auto_module_class = Scale
    _index_state = IdxType.Scalar

    def __init__(self, name, parents, func=torch.sqrt, module="auto", **kwargs):
        self.module_kwargs = {"func": func}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_MolMaskNode(AutoKw, SingleNode):
    _input_names = "notconverged"
    _auto_module_class = SEQM_MolMask
    _index_state = IdxType.Molecules

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = (parents,)
        self.module_kwargs = {}
        super().__init__(name, parents, module=module, **kwargs)


class AtomMaskNode(AutoKw, SingleNode):
    _input_names = "species"
    _auto_module_class = AtomMask
    _index_state = IdxType.Molecules

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = (parents,)
        self.module_kwargs = {}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_OrbitalMaskNode(AutoKw, SingleNode):
    _input_names = "species"
    _auto_module_class = SEQM_OrbitalMask
    _index_state = IdxType.Molecules

    def __init__(self, name, parents, target_method, nOccVirt=None, module="auto", **kwargs):
        parents = (parents,)
        self.module_kwargs = {"target_method": target_method, "nOccVirt": nOccVirt}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_MaskOnMolNode(AutoKw, SingleNode):
    """
    used for quantity like total energy, Heat of formation
    with shape (molecules,)
    """

    _input_names = "var", "mol_mask"
    _auto_module_class = SEQM_MaskOnMol
    _index_state = IdxType.NotFound

    def __init__(self, name, parents, module="auto", **kwargs):
        self.module_kwargs = {}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_MaskOnMolAtomNode(AutoKw, SingleNode):
    """
    used for quantity like force
    with shape (molecules,atoms)
    """

    _input_names = "var", "mol_mask", "atom_mask"
    _auto_module_class = SEQM_MaskOnMolAtom
    _index_state = IdxType.NotFound

    def __init__(self, name, parents, module="auto", **kwargs):
        self.module_kwargs = {}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_MaskOnMolOrbitalNode(AutoKw, SingleNode):
    """
    used for quantity like orbital energy
    with shape (molecules,orbitals)
    """

    _input_names = "var", "mol_mask", "orbital_mask"
    _auto_module_class = SEQM_MaskOnMolOrbital
    _index_state = IdxType.NotFound

    def __init__(self, name, parents, module="auto", **kwargs):
        self.module_kwargs = {}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_MaskOnMolOrbitalAtomNode(AutoKw, SingleNode):
    """
    used for quantity like orbital charge densit
    with shape (molecules,orbitals, atoms)
    """

    _input_names = "var", "mol_mask", "orbital_mask", "atom_mask"
    _auto_module_class = SEQM_MaskOnMolOrbitalAtom
    _index_state = IdxType.NotFound

    def __init__(self, name, parents, module="auto", **kwargs):
        self.module_kwargs = {}
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_EnergyNode(ExpandParents, AutoKw, MultiNode):
    _input_names = "par_atom", "Positions", "Species"
    _output_names = "mol_energy", "Etot_m_Eiso", "notconverged"
    _main_output = "mol_energy"
    _output_index_states = (IdxType.Molecules,) * len(_output_names)
    _auto_module_class = SEQM_Energy

    @_parent_expander.match(Network)
    def expand0(self, network, seqm_parameters, decay_factor=1.0e-2, **kwargs):

        n_target_peratom = len(seqm_parameters["learned"])

        par_atom = HChargeNode(
            "SEQM_Atom_Params", network, module_kwargs=dict(n_target=n_target_peratom, first_is_interacting=True)
        )

        with torch.no_grad():
            for layer in par_atom.torch_module.layers:
                layer.weight.data *= decay_factor
                layer.bias.data *= decay_factor

        positions = find_unique_relative(network, PositionsNode)
        species = find_unique_relative(network, SpeciesNode)

        return par_atom.main_output, positions, species

    _parent_expander.assertlen(3)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, None, None)

    def __init__(self, name, parents, seqm_parameters, decay_factor=1.0e-2, module="auto", **kwargs):
        parents = self.expand_parents(parents, seqm_parameters=seqm_parameters, decay_factor=decay_factor, **kwargs)
        self.module_kwargs = dict(seqm_parameters=seqm_parameters)
        super().__init__(name, parents, module=module, **kwargs)


class SEQM_AllNode(SEQM_EnergyNode):
    _input_names = "par_atom", "Positions", "Species"
    _output_names = (
        "mol_energy",
        "Etot_m_Eiso",
        "orbital_energies",
        "single_particle_density_matrix",
        "electric_energy",
        "nuclear_energy",
        "isolated_atom_energy",
        "orbital_charges",
        "notconverged",
        "atomic_charge",
    )
    _main_output = "mol_energy"
    _output_index_states = (IdxType.Molecules,) * len(_output_names)
    _auto_module_class = SEQM_All



"""
Layers for encoding, decoding, index states, besides pairs
"""

import warnings
import torch


class OneHotSpecies(torch.nn.Module):
    """
    Encodes species as one-hot map using `species_set`
    :param species_set iterable of species types. Typically this will be the Z-value of the elemeent.
    Note: 0 denotes a 'blank atom' used only for padding purposes

    :returns OneHotSpecies

    """

    def __init__(self, species_set):
        super().__init__()

        self.species_set = torch.as_tensor(species_set)
        self.n_species = self.species_set.shape[0]

        max_z = torch.max(self.species_set)
        zmap = torch.zeros((max_z + 1,), dtype=torch.long)
        zmap[self.species_set] = torch.arange(0, self.n_species, dtype=torch.long)

        self.species_map = torch.nn.Parameter(torch.as_tensor(zmap, dtype=torch.long), requires_grad=False)

    def forward(self, species):
        """
        :param species:
        :return: Initial one-hotted features, nonblank atoms
        """

        onehot_species = torch.eye(self.n_species, dtype=torch.bool, device=species.device)[self.species_map[species]]
        nonblank = ~onehot_species[:, :, 0]
        initial_features = onehot_species[:, :, 1:]  # remove atoms that are 0 in the species map.

        return initial_features, nonblank


class PaddingIndexer(torch.nn.Module):
    """
    Hipnn's indexer

    Description:
        This indexer allows us to go from rectangular (mol,atom) representations
        To (flatatom) representations and a corresponding
        molecule index for those atoms.
        The 'real_index' allows us to take _values_ from a flattened rectangular
        representation (mol*atom)
        And select only the real ones.
        The 'inv_real_index', when indexed, converts a (mol*atom) index set into a (flatatom) index set

    """

    def forward(self, features, nonblank):
        """
        Pytorch Enforced Forward function

        :param features:
        :param nonblank:
        :return: real atoms, amd inverse atoms mappings
        """
        dev = features.device
        n_molecules, n_atoms_max = nonblank.shape
        n_fictitious_atoms = nonblank.shape[0] * nonblank.shape[1]
        # Just calculate the total number of atoms in the dataset

        flat_nonblank = nonblank.reshape(n_fictitious_atoms)
        # Flatten the nonblank n_mol x n_atoms nonblank matrix

        large_enough = torch.empty(nonblank.shape,device=dev,dtype=torch.int64)
        large_enough.resize_(0)
        real_atoms = torch.nonzero(flat_nonblank, as_tuple=False,out=large_enough)[:, 0]
        # print(real_atoms)
        # Grab the indexes of each real atom, give it an atom get an index back

        n_real_atoms = real_atoms.shape[0]
        # Count how many real atoms there are

        inv_real_atoms = torch.zeros((n_fictitious_atoms,), dtype=torch.long, device=dev)
        # Create a vector of 0's

        inv_real_atoms[real_atoms] = torch.arange(n_real_atoms, dtype=torch.long, device=dev)
        # Create the inverse real atom "function" give it an index get an atom back

        # Flatten incoming features to atom representation
        indexed_features = features.reshape(n_molecules * n_atoms_max, -1)[real_atoms]
        if indexed_features.ndimension() == 1:
            indexed_features = indexed_features.unsqueeze(1)

        # Get molecule index for atoms
        mol_index_shaped = (
            torch.arange(n_molecules, dtype=torch.long, device=dev).unsqueeze(1).expand(-1, n_atoms_max)
        )
        atom_index_shaped = (
            torch.arange(n_atoms_max, dtype=torch.long, device=dev).unsqueeze(0).expand(n_molecules, -1)
        )
        atom_index = atom_index_shaped.reshape(n_fictitious_atoms)[real_atoms]
        mol_index = mol_index_shaped.reshape(n_fictitious_atoms)[real_atoms]

        return indexed_features, real_atoms, inv_real_atoms, mol_index, atom_index, n_molecules, n_atoms_max


class AtomReIndexer(torch.nn.Module):
    def forward(self, molatom_thing, real_atoms):
        m, a, *rest = molatom_thing.shape
        out = molatom_thing.reshape(m * a, *rest)[real_atoms]
        if len(rest) == 0:
            out = out.unsqueeze(1)
        return out


class MolSummer(torch.nn.Module):
    """
    Molecule Summer

    Description:
        This sums (flatatom) things into (mol) things.
        It actually works similarly to the interaction layer
    """

    def forward(self, features, mol_index, n_molecules):
        featshape = (1,) if features.ndimension() == 1 else features.shape[1:]
        out_shape = (n_molecules, *featshape)
        result = torch.zeros(*out_shape, device=features.device, dtype=features.dtype)
        result.index_add_(0, mol_index, features)

        return result


class SysMaxOfAtoms(torch.nn.Module):
    """
    Take maximum over atom dimension.
    """
    def forward(self, features, mol_index, n_molecules):
        # Add feature dimension if not found
        if features.ndim == 1:
            featshape = (1,)
            features = features.unsqueeze(1)
        else:
            featshape = features.shape[1:]
        # Allocate result
        out_shape = (n_molecules, *featshape)
        result = torch.zeros(*out_shape, device=features.device, dtype=features.dtype)

        # Prepare index shape for scatter operation
        mi_expand = mol_index.reshape(-1, *(1,) * len(featshape))
        mi_expand = mi_expand.expand((-1, *featshape))

        # Perform calculation
        result.scatter_reduce_(0, mi_expand, features, reduce='amax', include_self=False)
        return result

class AtomDeIndexer(torch.nn.Module):
    def forward(self, features, mol_index, atom_index, n_molecules, n_atoms_max):
        featshape = 1 if features.ndimension() == 1 else features.shape[1:]
        out_shape = (n_molecules, n_atoms_max, *featshape)
        result = torch.zeros(*out_shape, device=features.device, dtype=features.dtype)
        result[mol_index, atom_index] = features
        return result


class CellScaleInducer(torch.nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pbc = False

    def forward(self, coordinates, cell):
        strain = torch.eye(
            coordinates.shape[2], dtype=coordinates.dtype, device=coordinates.device, requires_grad=True
        ).unsqueeze(0)
        strained_coordinates = torch.bmm(coordinates, strain)
        if cell.dim() == 2:
            strained_cell = torch.mm(cell, strain.squeeze(0))
        return strained_coordinates, strained_cell, strain


class QuadPack(torch.nn.Module):
    """
    Converts quadrupoles flattened form to packed triangular, assumes symmmetric.
    Packed form: (molecule)   XX, YY, ZZ, XY, XZ, YZ
    Index                      0,  4,  8,  1,  2,  5
    Unpacked form: (molecule) XX, XY, XZ, YX, YY, YZ , ZX, ZY, ZZ
    Index                      00 01  02, 10, 11, 12,  20, 21, 22
    """

    def __init__(self):
        super().__init__()
        ind1 = [0, 1, 2, 0, 0, 1]
        ind2 = [0, 1, 2, 1, 2, 2]
        self.register_buffer("ind_1", torch.LongTensor(ind1))
        self.register_buffer("ind_2", torch.LongTensor(ind2))

    def forward(self, quadrupoles):
        return quadrupoles[:, self.ind1, self.ind2]


class QuadUnpack(torch.nn.Module):
    """
    Converts quadrupoles from packed triangular form to flattened molecule form.
    Packed form: (molecule)   XX, YY, ZZ, XY, XZ, YZ
    Index                      0,  1,  2,  3,  4,  5
    Unpacked form: (molecule) XX, XY, XZ, YX, YY, YZ , ZX, ZY, ZZ
    Index                      0,  3,  4,  3,  1,  5,  4,  5,  2
    """

    def __init__(self):
        super().__init__()
        indices = [0, 3, 4, 3, 1, 5, 4, 5, 2]
        self.register_buffer("index_permutation", torch.LongTensor(indices))

    def forward(self, packed_quadrupoles):
        return packed_quadrupoles[:, self.index_permutation]


class FilterBondsOneway(torch.nn.Module):
    def forward(self, bonds, pair_first, pair_second):
        # in seqm, only bonds with index first < second is used
        cond = pair_first < pair_second
        return bonds[cond]

class FuzzyHistogram(torch.nn.Module):
    """ 
    Transforms a scalar feature into a vectorized feature via 
    the fuzzy/soft histogram method.

    :param length: length of vectorized feature

    :returns FuzzyHistogram
    """

    def __init__(self, length, vmin, vmax):
        super().__init__()

        err_msg = "The value of 'length' must be a positive integer."
        if not isinstance(length, int):
            raise ValueError(err_msg)
        if length <= 0:
            raise ValueError(err_msg)

        if not (isinstance(vmin, (int,float)) and isinstance(vmax, (int,float))):
            raise ValueError("The values of 'vmin' and 'vmax' must be floating point numbers.")
        if vmin >= vmax:
            raise ValueError("The value of 'vmin' must be less than the value of 'vmax.'")

        self.bins = torch.nn.Parameter(torch.linspace(vmin, vmax, length), requires_grad=False)
        self.sigma = (vmax - vmin) / length

        self.vmin = vmin
        self.vmax = vmax

    def forward(self, values):
        # Warn user if provided values lie outside the range of the histogram bins
        values_out_of_range = (values < self.vmin) + (values > self.vmax)

        if values_out_of_range.sum() > 0:
            perc_out_of_range = values_out_of_range.float().mean()
            warnings.warn(
                "Values out of range for FuzzyHistogrammer\n"
                f"Number of values out of range: {values_out_of_range.sum()}\n"
                f"Percentage of values out of range: {perc_out_of_range * 100:.2f}%\n"
                f"Set range for FuzzyHistogrammer: ({self.vmin:.2f}, {self.vmax:.2f})\n"
                f"Range of values: ({values.min().item():.2f}, {values.max().item():.2f})"
            )

        if values.shape[-1] != 1:
            values = values[...,None]
        x = values - self.bins
        histo = torch.exp(-((x / self.sigma) ** 2) / 4)
        return torch.flatten(histo, end_dim=1)